@@ -65,12 +65,13 @@ static const aop_t acl_op_table[] = {
     { 's',      ACL_SETKEY },
     { 'x',      ACL_ALL_MASK },
     { '*',      ACL_ALL_MASK },
+    { 'e',      ACL_EXTRACT },
     { '\0',     0 }
 };
 
 typedef struct _wildstate {
-    int         nwild;
-    krb5_data   *backref[9];
+    int nwild;
+    const krb5_data *backref[9];
 } wildstate_t;
 
 static aent_t   *acl_list_head = (aent_t *) NULL;
@@ -93,7 +94,6 @@ static const char *acl_syn_err_msg = N_("%s: syntax error at line %d "
                                         "<%10s...>");
 static const char *acl_cantopen_msg = N_("%s while opening ACL file %s");
 
-
 /*
  * kadm5int_acl_get_line() - Get a line from the ACL file.
  *                      Lines ending with \ are continued on the next line
@@ -157,7 +157,7 @@ kadm5int_acl_get_line(fp, lnp)
     else
         return(acl_buf);
 }
-
+
 /*
  * kadm5int_acl_parse_line() - Parse the contents of an ACL line.
  */
@@ -260,7 +260,7 @@ kadm5int_acl_parse_line(lp)
            ("X kadm5int_acl_parse_line() = %x\n", (long) acle));
     return(acle);
 }
-
+
 /*
  * kadm5int_acl_parse_restrictions() - Parse optional restrictions field
  *
@@ -360,7 +360,7 @@ kadm5int_acl_parse_restrictions(s, rpp)
             code, (*rpp) ? (*rpp)->mask : 0));
     return code;
 }
-
+
 /*
  * kadm5int_acl_impose_restrictions()   - impose restrictions, modifying *recp, *maskp
  *
@@ -375,7 +375,7 @@ kadm5int_acl_impose_restrictions(kcontext, recp, maskp, rp)
     restriction_t              *rp;
 {
     krb5_error_code     code;
-    krb5_int32          now;
+    krb5_timestamp      now;
 
     DPRINT(DEBUG_CALLS, acl_debug_level,
            ("* kadm5int_acl_impose_restrictions(..., *maskp=0x%08x, rp=0x%08x)\n",
@@ -408,13 +408,14 @@ kadm5int_acl_impose_restrictions(kcontext, recp, maskp, rp)
     }
     if (rp->mask & KADM5_PRINC_EXPIRE_TIME) {
         if (!(*maskp & KADM5_PRINC_EXPIRE_TIME)
-            || (recp->princ_expire_time > (now + rp->princ_lifetime)))
+            || ts_after(recp->princ_expire_time,
+                        ts_incr(now, rp->princ_lifetime)))
             recp->princ_expire_time = now + rp->princ_lifetime;
         *maskp |= KADM5_PRINC_EXPIRE_TIME;
     }
     if (rp->mask & KADM5_PW_EXPIRATION) {
         if (!(*maskp & KADM5_PW_EXPIRATION)
-            || (recp->pw_expiration > (now + rp->pw_lifetime)))
+            || ts_after(recp->pw_expiration, ts_incr(now, rp->pw_lifetime)))
             recp->pw_expiration = now + rp->pw_lifetime;
         *maskp |= KADM5_PW_EXPIRATION;
     }
@@ -434,7 +435,7 @@ kadm5int_acl_impose_restrictions(kcontext, recp, maskp, rp)
            ("X kadm5int_acl_impose_restrictions() = 0, *maskp=0x%08x\n", *maskp));
     return 0;
 }
-
+
 /*
  * kadm5int_acl_free_entries() - Free all ACL entries.
  */
@@ -468,7 +469,7 @@ kadm5int_acl_free_entries()
     acl_inited = 0;
     DPRINT(DEBUG_CALLS, acl_debug_level, ("X kadm5int_acl_free_entries()\n"));
 }
-
+
 /*
  * kadm5int_acl_load_acl_file() - Open and parse the ACL file.
  */
@@ -541,17 +542,15 @@ kadm5int_acl_load_acl_file()
            ("X kadm5int_acl_load_acl_file() = %d\n", retval));
     return(retval);
 }
-
+
 /*
  * kadm5int_acl_match_data()    - See if two data entries match.
  *
  * Wildcarding is only supported for a whole component.
  */
 static krb5_boolean
-kadm5int_acl_match_data(e1, e2, targetflag, ws)
-    krb5_data   *e1, *e2;
-    int         targetflag;
-    wildstate_t *ws;
+kadm5int_acl_match_data(const krb5_data *e1, const krb5_data *e2,
+                        int targetflag, wildstate_t *ws)
 {
     krb5_boolean        retval;
 
@@ -589,15 +588,13 @@ kadm5int_acl_match_data(e1, e2, targetflag, ws)
     DPRINT(DEBUG_CALLS, acl_debug_level, ("X acl_match_entry()=%d\n",retval));
     return(retval);
 }
-
+
 /*
  * kadm5int_acl_find_entry()    - Find a matching entry.
  */
 static aent_t *
-kadm5int_acl_find_entry(kcontext, principal, dest_princ)
-    krb5_context        kcontext;
-    krb5_principal      principal;
-    krb5_principal      dest_princ;
+kadm5int_acl_find_entry(krb5_context kcontext, krb5_const_principal principal,
+                        krb5_const_principal dest_princ)
 {
     aent_t              *entry;
     krb5_error_code     kret;
@@ -697,7 +694,7 @@ kadm5int_acl_find_entry(kcontext, principal, dest_princ)
     DPRINT(DEBUG_CALLS, acl_debug_level, ("X kadm5int_acl_find_entry()=%x\n",entry));
     return(entry);
 }
-
+
 /*
  * kadm5int_acl_init()  - Initialize ACL context.
  */
@@ -720,7 +717,7 @@ kadm5int_acl_init(kcontext, debug_level, acl_file)
     DPRINT(DEBUG_CALLS, acl_debug_level, ("X kadm5int_acl_init() = %d\n", kret));
     return(kret);
 }
-
+
 /*
  * kadm5int_acl_finish  - Terminate ACL context.
  */
@@ -733,7 +730,7 @@ kadm5int_acl_finish(kcontext, debug_level)
     kadm5int_acl_free_entries();
     DPRINT(DEBUG_CALLS, acl_debug_level, ("X kadm5int_acl_finish()\n"));
 }
-
+
 /*
  * kadm5int_acl_check_krb()     - Is this operation permitted for this principal?
  */
